import sys
import os
import json
from copy import deepcopy

from typing import (
    Dict,
    Any,
)

import numpy as np
import scipy as sp
from scipy.linalg import svdvals, expm

from . import accessories as acc
from .accessories import (
    FOCArray,
    FocontError,
)


PData = Dict[str, Any]


def _parse_matrix_definition(d: str, n: int) -> FOCArray | None:
    if type(d) == str:
        d = d.strip()
        ind = d.find("I")
        if ind > -1:
            d = d.strip("I").strip()
            if d:
                g = float(d)
            else:
                g = 1.0

            return g * np.eye(n)

    return None


# TODO: Check if B and C are full rank
def _validate_input(pdata: PData) -> None:
    # ** A
    if "A" not in pdata:
        raise RuntimeError("System matrix 'A' is not defined.")
    else:
        n = len(pdata["A"])
        if n < 1:
            raise RuntimeError("System matrix 'A' has dimension zero.")
        else:
            if n != len(pdata["A"][0]):
                raise RuntimeError("System matrix 'A' is not square.")

    pdata["A"] = np.array(pdata["A"])

    # ** B
    if "B" not in pdata:
        raise RuntimeError("System matrix 'B' is not defined.")
    else:
        n1 = len(pdata["B"])
        if n1 != n:
            raise RuntimeError(
                "System matrices 'A' and 'B' must " "have the same number of rows"
            )
        m = len(pdata["B"][0])
        if m < 1:
            raise RuntimeError("System matrix 'B' has dimension zero.")

    pdata["B"] = np.array(pdata["B"])

    # ** C
    if "C" not in pdata:
        raise RuntimeError("System matrix 'C' is not defined.")
    elif type(pdata["C"]) == list:
        r = len(pdata["C"])
        if r < 1:
            raise RuntimeError("System matrix 'C' has dimension zero.")
        n2 = len(pdata["C"][0])
        if n2 != n:
            raise RuntimeError(
                "System matrices 'A' and 'C' must " "have the same number of columns."
            )

        pdata["C"] = np.array(pdata["C"])

    elif type(pdata["C"]) == str:
        try:
            pdata["C"] = _parse_matrix_definition(pdata["C"], n)
        except:
            raise RuntimeError(
                "Matrix 'C={}' is an invalid definition.".format(pdata["C"])
            )

    # ** Q
    if "Q" not in pdata:
        raise RuntimeError("Cost function weight 'Q' is not defined.")
    elif type(pdata["Q"]) == list:
        n3 = len(pdata["Q"])
        if n3 < 1:
            raise RuntimeError("Cost function weight 'Q' has dimension zero.")
        elif n3 != n:
            raise RuntimeError(
                "Matrices 'A' and 'Q' must " "have the same number of columns and rows."
            )
        n4 = len(pdata["Q"][0])
        if n3 != n4:
            raise RuntimeError("Cost function weight 'Q' must be a square matrix.")

        pdata["Q"] = np.array(pdata["Q"])
    elif type(pdata["Q"]) == str:
        if pdata["Q"] == "default":
            pdata["Q"] = np.matmul(pdata["C"].T, pdata["C"])
        else:
            try:
                pdata["Q"] = _parse_matrix_definition(pdata["Q"], n)
            except:
                raise RuntimeError(
                    "Matrix 'Q={}' has an invalid definition.".format(pdata["Q"])
                )

    if not acc.is_symmetric(pdata["Q"]):
        raise RuntimeError("Cost function weight 'Q' is not symmetric.")

    # ** R
    if "R" not in pdata:
        raise RuntimeError("Cost function weight 'R' is not defined.")
    elif type(pdata["R"]) == list:
        n5 = len(pdata["R"])
        if n5 < 1:
            raise RuntimeError("Cost function weight 'R' has dimension zero.")
        elif n5 != m:
            raise RuntimeError(
                "Matrices 'B' and 'R' must " "have the same number of columns."
            )
        n6 = len(pdata["R"][0])
        if n5 != n6:
            raise RuntimeError("Cost function weight 'R' must be a square matrix.")

        pdata["R"] = np.array(pdata["R"])
    elif type(pdata["R"]) == str:
        try:
            pdata["R"] = _parse_matrix_definition(pdata["R"], m)
        except:
            raise RuntimeError(
                "Matrix 'R={}' has an invalid definition.".format(pdata["R"])
            )

    if not acc.is_symmetric(pdata["R"]):
        raise RuntimeError("Cost function weight 'R' is not symmetric.")

    # ** Q
    if "Q0" not in pdata:
        raise RuntimeError("Cost function weight 'Q0' is not defined.")
    elif type(pdata["Q0"]) == list:
        n3 = len(pdata["Q0"])
        if n3 < 1:
            raise RuntimeError("Cost function weight 'Q0' has dimension zero.")
        elif n3 != n:
            raise RuntimeError(
                "Matrices 'A' and 'Q0' must "
                "have the same number of columns and rows."
            )
        n4 = len(pdata["Q0"][0])
        if n3 != n4:
            raise RuntimeError("Cost function weight 'Q0' must be a square matrix.")

        pdata["Q0"] = np.array(pdata["Q0"])
    elif type(pdata["Q0"]) == str:
        if pdata["Q0"] == "default":
            C = pdata["C"]
            pdata["Q0"] = np.eye(C.shape[1])
        else:
            try:
                pdata["Q0"] = _parse_matrix_definition(pdata["Q0"], n)
            except:
                raise RuntimeError(
                    "Matrix 'Q0={}' has an invalid definition.".format(pdata["Q0"])
                )

    if not acc.is_symmetric(pdata["Q0"]):
        raise RuntimeError("Cost function weight 'Q0' is not symmetric.")

    # ** type
    if "type" not in pdata:
        raise RuntimeError(
            "System type must be set as discrete (D) " "or continuous (C)."
        )
    elif not (pdata["type"] == "C" or pdata["type"] == "D"):
        raise RuntimeError(
            "System type must be set as discrete (D) " "or continuous (C)."
        )


def _discretize(pdata: PData) -> None:
    A: FOCArray = pdata["A"]
    B: FOCArray = pdata["B"]
    Ts: FOCArray = pdata["Ts"]
    TsA: FOCArray = Ts * A
    Ad: FOCArray = expm(TsA)

    n: int = A.shape[0]
    rankA: int = np.linalg.matrix_rank(A)

    if rankA == n:
        I = np.eye(n)
        Ainv = np.linalg.inv(A)
        Mb = np.matmul(Ainv, Ad - I)

        Bd = np.matmul(Mb, B)
    else:
        zoh_calc_step = pdata["zoh_calc_step"]
        zoh_Ts = Ts / zoh_calc_step

        Mb = np.eye(n)
        tau = zoh_Ts
        while tau < Ts:
            Atau = expm(tau * A)
            Mb += zoh_Ts * Atau

            tau += zoh_Ts

        Bd = np.matmul(Mb, B)

    pdata["Ac"] = A
    pdata["Bc"] = B

    pdata["A"] = Ad
    pdata["B"] = Bd

    pdata["Aplant_discretized"] = Ad
    pdata["Bplant_discretized"] = Bd


def _fo_controller_structure(pdata: PData) -> None:
    from scipy.linalg import block_diag

    A: FOCArray = pdata["A"]
    B: FOCArray = pdata["B"]
    C: FOCArray = pdata["C"]
    Q: FOCArray = pdata["Q"]
    R: FOCArray = pdata["R"]
    Q0: FOCArray = pdata["Q0"]

    nc: int = pdata["controller_order"]
    Ccont: FOCArray = pdata["Ccont"]
    Dcont: FOCArray = pdata["Dcont"]
    Qcont: FOCArray = pdata["Qcont"]
    Rcont: FOCArray = pdata["Rcont"]
    Q0cont: FOCArray = pdata["Q0cont"]

    n: int = A.shape[0]
    m: int = B.shape[1]
    r: int = C.shape[0]

    A11: FOCArray = A + B @ Dcont @ C
    A12: FOCArray = B @ Ccont

    Aext: FOCArray = np.block([[A11, A12], [np.zeros((nc, n + nc))]])
    Bext: FOCArray = np.block([[np.zeros((n, nc))], [np.eye(nc)]])
    Cext: FOCArray = np.block([[np.zeros((nc, n)), np.eye(nc)], [C, np.zeros((r, nc))]])
    Qext: FOCArray = block_diag(Q, Qcont)
    Q0ext: FOCArray = block_diag(Q0, Q0cont)
    Rext: FOCArray = block_diag(Rcont)

    pdata["A"] = Aext
    pdata["B"] = Bext
    pdata["C"] = Cext
    pdata["Q"] = Qext
    pdata["R"] = Rext
    pdata["Q0"] = Q0ext


def _controller_structure(pdata: PData) -> None:
    s: str = pdata["structure"]

    if s == "FO":
        m = pdata["B"].shape[1]
        if "controller_order" not in pdata:
            pdata["controller_order"] = m

        nc = pdata["controller_order"]

        if "Ccont" not in pdata:
            Ccont = np.eye(m)
            if nc > m:
                Ccont = np.block([Ccont, np.zeros((m, nc - m))])

            pdata["Ccont"] = Ccont

        rC, cC = pdata["Ccont"].shape
        if cC != nc or rC != m:
            raise FocontError("The dimension of 'Ccont' must be {}x{}.".format(m, nc))

        r = pdata["C"].shape[0]
        if "Dcont" not in pdata:
            pdata["Dcont"] = np.zeros((m, r))

        rD, cD = pdata["Dcont"].shape
        if rD != m and cD != r:
            raise FocontError("The dimension of 'Dcont' must be {}x{}.".format(m, r))

        for mtx in ["Qcont", "Q0cont", "Rcont"]:
            if mtx not in pdata:
                pdata[mtx] = np.eye(nc)
            elif type(pdata[mtx]) == str:
                pdata[mtx] = _parse_matrix_definition(pdata[mtx], nc)

        _fo_controller_structure(pdata)
    elif s == "SOF":
        pass
    else:
        raise FocontError("Undefined controller structure '{}'.".format(s))


def load_from_json_file(json_filename: str) -> PData:
    with open(json_filename, "r") as fp:
        jobj: PData = json.load(fp)

    return jobj


def load_from_mat_file(filename: str) -> PData:
    from scipy.io import loadmat

    mat: PData = loadmat(filename)

    for var_name in mat:
        if hasattr(mat[var_name], "dtype") and str(mat[var_name].dtype).startswith(
            "<U"
        ):
            mat[var_name] = str(mat[var_name][0])
        elif var_name == "controller_order":
            mat[var_name] = int(mat[var_name].item(0))

    return mat


def load(input_data: PData | str) -> PData:
    r"""
    Load Fixed Order Controller problem paramters from
    a Python data structure or from a json, or mat file.

    :arg input_data dict_or_str: The source from which the problem
            parameters will be loaded.

    `input_data` can be json or mat filepath. In this case, file will
    be read and problem parameters data structre will be created from
    the json or mat file.

    *NOTE*: Matrices must be Python array of array of floats with appropriate row
    and column sizes (They are not `numpy` arrays!). Some matrices can be defined
    as a string for ease of use. E.g:
        `C = 'I'` or `Q = '1e-2I'`

        They will be translated to `numpy` identity matrices, `np.eye(n)` and
        `1e-2 * np.eye(n)`, where `n` is the dimension of LTI systems state vector.

    `focont` expects the following paramters:
        * `A`: System matrix of the LTI sytem ($A \in \mathbb{R}_{n \times n}$).
        * `B`: Input matrix ($B \in \mathbb{R}_{n \times m}$).
        * `C`: Output matrix (can be defined as a string, see the note above.)
                ($C \in \mathbb{R}_{r \times n}$).

        * `Q` (optional): Cost function weight for LTI system states (can be defined as a string.)
        `Q` must have the same dimension as `A` and must be symmetric and
        semi-positive definite. If it is not provided, its default value is 'I'.
        * `R` (optional): Cost function weight for LTI system's input. (can be a string.)
        `R` must be square and have the same number of columns as `B`. It must be
        symmetric and positive definite. Its default value is 'I'.
        * `Q0' (optional): It has the same properties as `Q`, but it is used for
        calculating an appropriate realization of the LTI system as an intermediary
        step of the algorithm. Its default value is 'I'.
        * `type` (optional): It can be 'D' if the LTI system is discrete and
        'C' if it is continuous. Its default value is 'D'.
        * `Ts` (optional): It is the sampling period used for ZOH discretization of
        the LTI system. Its default value is '0.01'.
        * `max_iter` (optional): Dynamic programming iterations limit. Its default
        value is '1e6'.
        * `eps_conv` (optional): Condition for convergence. If change in the cost-to-go
        function is smaller than this value, iterations will be terminated. Its
        default value is '1e-12'.
        * `zoh_calc_step` (optional): Max number of iterations used in ZOH discretization. Its
        default value is '256'.
        * `structure` (optional): It is 'SOF', if a static output feedback is wanted
        to be calculated. It is 'FO' if controller is dynamic.

        If controller structre is dynamic, then the paramters below can be provided.
            * `Ccont` (optional): Output matrix of the proposed dynamic controller.
            Its default value is '$I_{m \times m}$'.
            * `Dcont` (optional): Input to output gain of the controller. Its
            default value is '$0_{m \times r}$'.
            * `Qcont` (optional): Cost function weight on controller's state vector.
            Its default value is 'I'.
            * `Q0cont` (optional)
            * `Rcont` (optional): Cons function weight on controller's input vector.
            Its default value is 'I'.
    """
    pdata: PData = {}
    if type(input_data) == str:
        filename = input_data
        _, ext = os.path.splitext(filename)

        if ext == ".json":
            pdata = load_from_json_file(filename)
        elif ext == ".mat":
            pdata = load_from_mat_file(filename)
    else:
        assert isinstance(input_data, dict)
        pdata = deepcopy(input_data)

    if "type" not in pdata:
        pdata["type"] = "D"

    if "Q" not in pdata:
        pdata["Q"] = "default"

    if "R" not in pdata:
        pdata["R"] = "I"

    if "Q0" not in pdata:
        pdata["Q0"] = "default"

    if "Ts" not in pdata:
        pdata["Ts"] = 0.01

    if "max_iter" not in pdata:
        pdata["max_iter"] = int(1e6)
    else:
        pdata["max_iter"] = int(pdata["max_iter"])

    if "eps_conv" not in pdata:
        pdata["eps_conv"] = 1e-12

    if "zoh_calc_step" not in pdata:
        pdata["zoh_calc_step"] = 256

    if "structure" not in pdata:
        pdata["structure"] = "SOF"

    _validate_input(pdata)

    stable = acc.is_stable(pdata["type"], np.linalg.eigvals(pdata["A"]))
    pdata["open_loop_stable"] = stable

    pdata["Aplant"] = pdata["A"]
    pdata["Bplant"] = pdata["B"]
    pdata["Cplant"] = pdata["C"]
    pdata["Qplant"] = pdata["Q"]
    pdata["Rplant"] = pdata["R"]
    pdata["Q0plant"] = pdata["Q0"]

    if pdata["type"] == "C":
        _discretize(pdata)

    pdata["open_loop_lti"] = acc.convert_to_lti(pdata["A"], pdata["B"], pdata["C"])
    _controller_structure(pdata)

    pdata["pdata_initialized"] = True

    return pdata



import numpy as np
from scipy import signal

from typing import (
    List,
    TypeVar,
    Any,
    Optional,
    Tuple,
    Self,
)

from numpy.typing import (
    ArrayLike,
    NDArray,
)

# TODO: Make this type equivalent to ndarray dtype
DType = Any  # TypeVar("DType", bound=np.float64)
FOCArray = NDArray[DType]


def is_stable(type: str, evals: ArrayLike) -> bool:
    if type == "C":
        if np.any(np.real(evals) >= 0):
            return False
        else:
            return True
    elif type == "D":
        if np.any(np.abs(evals) >= 1):
            return False
        else:
            return True
    else:
        raise FocontError("Undefined system type '{}'.".format(type))


def is_symmetric(a: NDArray[DType], rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    return np.allclose(a, a.T, rtol=rtol, atol=atol)


def h2_norm(lti_mimo: Any) -> float:
    impulse_responses: List[List[NDArray[DType]]] = [[]]
    max_t: int = 0

    lti_siso = lti_mimo[0][0]
    is_D: bool = lti_siso.dt is not None

    r: int = len(lti_mimo)
    m: int = len(lti_mimo[0])

    h2: float = 0

    for i in range(r):
        for j in range(m):
            lti = lti_mimo[i][j]

            _, y = lti.impulse()

            # NOTE: `impulse` function implementations in SciPy
            # are different for discret and continuous LTI.
            if is_D:
                y = y[0]

            impulse_responses[i] += [y]

            if y.shape[0] > max_t:
                max_t = y.shape[0]

        impulse_responses += [[]]

    for t in range(max_t):
        y_mimo = np.zeros((r, m))

        for i in range(r):
            for j in range(m):
                resp = impulse_responses[i][j]
                if len(resp) > t:
                    yt = resp[t]
                    y_mimo[i, j] = yt.item(0)

        mag = np.linalg.norm(y_mimo)

        h2 += mag.item(0)

    return h2


def convert_to_lti(
    A: FOCArray,
    B: FOCArray,
    C: FOCArray,
    D: FOCArray = np.zeros((0, 0)),
    t: str = "D",
) -> Any:
    n: int = A.shape[0]
    m: int = B.shape[1]
    r: int = C.shape[0]

    if D.shape[0] == 0:
        D = np.zeros((r, m))

    if t == "D":
        lti_func = signal.dlti
    elif t == "C":
        lti_func = signal.lti

    lti: List[List[Any]] = [[]]
    for i in range(m):
        for j in range(r):
            lti_ij = lti_func(
                A, B[:, j : j + 1], C[i : i + 1, :], D[i : i + 1, j : j + 1]
            )

            lti[i] += [lti_ij]

        lti += [[]]

    if not lti[-1]:
        del lti[-1]

    return lti


def freq_response(
    A: FOCArray, B: FOCArray, C: FOCArray, D: FOCArray, N: int, xscale: str = "log"
) -> Tuple[FOCArray, FOCArray]:
    if xscale == "log":
        f = np.logspace(-3, 0, N)
    elif xscale == "lin":
        f = np.linspace(0, 1, N)

    n: int = A.shape[0]
    m: int = B.shape[1]
    r: int = C.shape[0]

    om: FOCArray = np.pi * f
    z: FOCArray = np.exp(1j * om)
    I: FOCArray = np.eye(n)

    resp: FOCArray = np.zeros((r, m, N)) + 1j * np.zeros((r, m, N))

    i: int = 0
    for zi in z:
        IA = zi * I - A

        IAinv = np.linalg.inv(IA)

        R = np.matmul(C, np.matmul(IAinv, B)) + D

        resp[:, :, i : i + 1] = R

        i += 1

    return f, resp


def message(msg: str, indent: int = 0) -> None:
    print("-" * indent + " " + msg)


def warning(msg: str, indent: int = 0) -> None:
    print("-" * indent + " WARNING: " + msg)


class FocontError(Exception):
    """General exception class for focont."""

    def __init__(self: Self, message: str = "An error occured.") -> None:
        self.message = message
        super(FocontError, self).__init__(self.message)



import sys
import os
import json

from typing import (
    Any,
    List,
    Tuple,
)

import numpy as np
import scipy as sp
from scipy.linalg import svdvals
from scipy import signal

from .accessories import (
    FocontError,
    message,
    is_stable,
    warning,
    h2_norm,
    convert_to_lti,
    freq_response,
    FOCArray,
)

from .system import PData


def _adequate_real(pdata: PData) -> None:
    A: FOCArray = pdata["A"]
    B: FOCArray = pdata["B"]
    C: FOCArray = pdata["C"]
    Q: FOCArray = pdata["Q"]
    Q0: FOCArray = pdata["Q0"]
    R0: FOCArray = np.eye(np.shape(C)[0])

    S0: FOCArray = sp.linalg.solve_discrete_are(A.T, C.T, Q0, R0)

    rankS0: int = np.linalg.matrix_rank(S0)
    if rankS0 != S0.shape[0]:
        raise RuntimeError("An adequate realization could not be found.")

    # S0 = v*e/v
    e: FOCArray
    v: FOCArray
    e, v = np.linalg.eigh(S0)
    e = np.sqrt(e)
    einv: FOCArray = np.divide(1.0, e)
    e = np.diag(e)
    einv = np.diag(einv)
    vinv: FOCArray = np.linalg.inv(v)

    # Similarity transformation
    Ta: FOCArray = v @ e @ vinv
    Tainv: FOCArray = v @ einv @ vinv

    Aa: FOCArray = Tainv @ A @ Ta
    Ba: FOCArray = Tainv @ B
    Ca: FOCArray = C @ Ta
    Qa: FOCArray = Ta.T @ Q @ Ta

    pdata["original_A"] = A
    pdata["original_B"] = B
    pdata["original_C"] = C
    pdata["original_Q"] = Q

    pdata["A"] = Aa
    pdata["B"] = Ba
    pdata["C"] = Ca
    pdata["Q"] = Qa
    pdata["Ta"] = Ta

    n: int = Aa.shape[0]
    I: FOCArray = np.eye(n)

    Cinv: FOCArray = np.linalg.pinv(Ca)
    Pi_c: FOCArray = Cinv @ Ca
    Pi_cbar: FOCArray = I - Pi_c
    Aa_cbar: FOCArray = Aa @ Pi_cbar

    svals: FOCArray = svdvals(Aa_cbar, overwrite_a=True)

    if np.any(svals >= 1):
        raise FocontError(
            "Projected system matrix has singular values "
            "greater than one\nfocont could not find a solution."
        )

    pdata["Pi_c"] = Pi_c
    pdata["Pi_cbar"] = Pi_cbar
    pdata["A_cbar"] = Aa_cbar
    pdata["Cinv"] = Cinv


def _closed_loop_FO_structure(pdata: PData) -> None:
    if pdata["type"] == "C":
        A = pdata["Aplant_discretized"]
        B = pdata["Bplant_discretized"]
    elif pdata["type"] == "D":
        A = pdata["Aplant"]
        B = pdata["Bplant"]

    C: FOCArray = pdata["Cplant"]

    m: int = B.shape[1]
    r: int = C.shape[0]

    # Controler dimension
    nc: int = pdata["controller_order"]

    K: FOCArray = pdata["K"]
    Acont: FOCArray = K[:, :nc]
    Bcont: FOCArray = K[:, nc:]
    Ccont: FOCArray = pdata["Ccont"]
    Dcont: FOCArray = pdata["Dcont"]

    A11: FOCArray = A + B @ Dcont @ C
    A12: FOCArray = B @ Ccont
    A21: FOCArray = Bcont @ C
    A22: FOCArray = Acont

    Acl: FOCArray = np.block([[A11, A12], [A21, A22]])

    ev_cl: FOCArray = np.linalg.eigvals(Acl)

    if not is_stable("D", ev_cl):
        raise FocontError(
            "Resulting fixed order controller does not stabilize "
            "the discrete time (or discretized) system."
        )

    Bcl: FOCArray = np.block([[B], [np.zeros((nc, m))]])

    Ccl: FOCArray = np.block([C, np.zeros((r, nc))])

    closed_loop_lti: Any = convert_to_lti(Acl, Bcl, Ccl)
    pdata["closed_loop_lti"] = closed_loop_lti

    controller_lti: Any = convert_to_lti(Acont, Bcont, Ccont, Dcont)
    pdata["controller_lti"] = controller_lti


def _closed_loop_system(pdata: PData) -> None:
    K: FOCArray = pdata["K"]

    A: FOCArray = pdata["original_A"]
    B: FOCArray = pdata["original_B"]
    C: FOCArray = pdata["original_C"]

    Acld: FOCArray = A + B @ K @ C
    ev_cl_d: FOCArray = np.linalg.eigvals(Acld)

    if not is_stable("D", ev_cl_d):
        raise FocontError(
            "Closed loop discrete time system matrix " "'A+BKC' is not stable."
        )

    pdata["Acld"] = Acld
    pdata["ev_cl_d"] = ev_cl_d

    if pdata["structure"] == "SOF":
        if pdata["type"] == "C":
            Ac = pdata["Ac"]
            Bc = pdata["Bc"]

            Aclc = Ac + np.matmul(Bc, np.matmul(K, C))
            ev_cl_c = np.linalg.eigvals(Aclc)

            if not is_stable(pdata["type"], ev_cl_c):
                raise FocontError(
                    "Closed loop continuous time system matrix "
                    "'Ac+BcKCc' is not stable. "
                    "Try to decrease sampling period."
                )

            pdata["Aclc"] = Aclc
            pdata["ev_cl_c"] = ev_cl_c

            m = Bc.shape[1]
            r = C.shape[0]

            closed_loop_lti: List[List[Any]] = [[]]
            for i in range(r):
                for j in range(m):
                    closed_loop_lti_ij = signal.lti(
                        Aclc, Bc[:, j : j + 1], C[i : i + 1, :], np.zeros((1, 1))
                    )

                    closed_loop_lti[i] += [closed_loop_lti_ij]

                closed_loop_lti += [[]]

            if not closed_loop_lti[-1]:
                del closed_loop_lti[-1]

            pdata["closed_loop_lti"] = closed_loop_lti
        elif pdata["type"] == "D":
            A = pdata["original_A"]
            B = pdata["original_B"]
            C = pdata["original_C"]

            Acl = A + np.matmul(B, np.matmul(K, C))

            m = B.shape[1]
            r = C.shape[0]

            closed_loop_lti = [[]]
            for i in range(r):
                for j in range(m):
                    closed_loop_lti_ij = signal.dlti(
                        Acl, B[:, j : j + 1], C[i : i + 1, :], np.zeros((1, 1))
                    )

                    closed_loop_lti[i] += [closed_loop_lti_ij]

                closed_loop_lti += [[]]

            if not closed_loop_lti[-1]:
                del closed_loop_lti[-1]

            pdata["closed_loop_lti"] = closed_loop_lti
    elif pdata["structure"] == "FO":
        _closed_loop_FO_structure(pdata)
    else:
        raise FocontError(
            "Unknown controller structure '{}'".format(pdata["structure"])
        )


def _calculate_sof(pdata: PData) -> None:
    A: FOCArray = pdata["A"]
    A_cbar: FOCArray = pdata["A_cbar"]
    B: FOCArray = pdata["B"]
    C: FOCArray = pdata["C"]
    Q: FOCArray = pdata["Q"]
    R: FOCArray = pdata["R"]
    Cinv: FOCArray = pdata["Cinv"]
    Pi_cbar: FOCArray = pdata["Pi_cbar"]

    max_iter: int = pdata["max_iter"]
    eps_conv: float = pdata["eps_conv"]

    progress: int = 10
    progress_step: int = max_iter // 10

    inaccurate_result: bool = False
    converged: bool = False
    P: FOCArray = np.copy(Q)

    for i in range(max_iter):
        P_pre = np.copy(P)

        Rbar = B.T @ P @ B + R
        Rinv = np.linalg.inv(Rbar)

        M1 = P @ B @ Rinv @ B.T @ P
        M2 = A.T @ (P - M1) @ A
        M3 = A_cbar.T @ M1 @ A_cbar

        P = Q + M2 + M3
        normP = np.linalg.norm(P)
        dP = np.linalg.norm(P - P_pre) / normP

        if dP < eps_conv:
            if np.isnan(dP) or np.isinf(dP):
                raise FocontError("Iterations did not converge.")
            else:
                message("Iterations converged, a solution is found")
                converged = True
                break

        if not inaccurate_result and normP * eps_conv > 1e2:
            warning("Cost-to-go is so large. Results can be inaccurate.")
            inaccurate_result = True

        if i % progress_step == 0:
            message("Progress:\t{}%, dP={}".format(progress, dP))
            progress += 10

    if not converged:
        raise FocontError(
            "Max iteration is reached but did not converge.\n"
            "Increase 'max_iter' or 'eps_conv' and try again."
        )

    F: FOCArray = -Rinv @ B.T @ P @ A
    K: FOCArray = F @ Cinv

    pdata["P"] = P
    pdata["F"] = F
    pdata["K"] = K

    _closed_loop_system(pdata)


def _create_open_loop_lti(pdata: PData) -> None:
    lti_func: Any
    if pdata["type"] == "C":
        lti_func = signal.lti
    elif pdata["type"] == "D":
        lti_func = signal.dlti

    A: FOCArray = pdata["Aplant"]
    B: FOCArray = pdata["Bplant"]
    C: FOCArray = pdata["Cplant"]

    r: int = C.shape[0]
    m: int = B.shape[1]

    open_loop_lti: List[List[Any]] = []
    for i in range(r):
        open_loop_lti += [[]]

        for j in range(m):
            open_loop_lti_ij = lti_func(
                A, B[:, j : j + 1], C[i : i + 1, :], np.zeros((1, 1))
            )

            open_loop_lti[i] += [open_loop_lti_ij]

    pdata["open_loop_lti"] = open_loop_lti


def _validate(pdata: PData) -> None:
    if "pdata_initialized" in pdata and pdata["pdata_initialized"]:
        pass
    else:
        raise FocontError(
            "Please, call 'system.load' function with "
            "your initial data set to generate problem "
            "data structure."
        )


def _invalidate(pdata: PData) -> None:
    pdata["pdata_initialized"] = False


def solve(pdata: PData) -> None:
    """
    Solves the SOF (static output feedback) or FOC (fixed order controller)
    problem for the given LTI (discrete or continous) system by applying the
    proposed solution method [1-2].

    [1]: Demir, O. and Özbay, H., 2020. Static output feedback stabilization
    of discrete time linear time invariant systems based on approximate dynamic
    programming. Transactions of the Institute of Measurement and Control,
    42(16), pp.3168-3182.

    [2]: Demir, O., 2020. Optimality based structured control of distributed
    parameter systems (Doctoral dissertation, Bilkent University).

    :arg pdata dict: Python dictionary of problem parameters obtained from
    `system.load` function of `focont` library.

    Controller is calculated by performing the following steps;

    1. Find an appropriate realization of the LTI system.
    2. Apply the approximate dyanmic programming (ADP) iterations to
    calculate the stabilizing controller which minimize a quadratic cost
    function similart to the well-known linear quadratic regulator (LQR)
    problem.

    *NOTE*: Solution is appended to the input argument `pdata`.
    """

    _validate(pdata)

    _create_open_loop_lti(pdata)
    _adequate_real(pdata)
    _calculate_sof(pdata)

    _invalidate(pdata)


def print_results(pdata: PData) -> None:
    with np.printoptions(precision=4):
        if pdata["structure"] == "SOF":
            message("Stabilizing SOF gain:", indent=1)
            print(pdata["K"])
            message("Eigenvalues of the closed loop system:", indent=1)
            if pdata["type"] == "C":
                print(pdata["ev_cl_c"])
            elif pdata["type"] == "D":
                print(pdata["ev_cl_d"])
                message("|e|:")
                print(np.abs(pdata["ev_cl_d"]))
        elif pdata["structure"] == "FO":
            nc = pdata["controller_order"]
            K = pdata["K"]
            Acont = K[:, :nc]
            Bcont = K[:, nc:]
            message("Acont:", indent=1)
            print(Acont)
            message("Bcont:", indent=1)
            print(Bcont)


def get_controller(pdata: PData, i: int = -1, j: int = -1) -> Any:
    """
    Returns the controller in SciPy discrete LTI system form.

    :arg pdata dict: Problem data structure.
    :arg i int: Controller output index for the MIMO controller.
    :arg j int: Controller input index for the MIMO controller.

    Returns an `m` by `r` Python array when `i` or `j` is not provided.
    The ith row and jth column of the return value gives the discrete LTI
    system from jth input to the ith output.
    """
    if i == -1 or j == -1:
        return pdata["controller_lti"]
    elif isinstance(pdata["controller_lti"], list):
        try:
            return pdata["controller_lti"][i][j]
        except IndexError:
            m = len(pdata["controller_lti"])
            if m > 0 and isinstance(pdata["controller_lti"][0], list):
                n = len(pdata["controller_lti"][0])
            else:
                return None
            raise FocontError(
                f"Controller dimension is {m}x{n}." f"Can not access ({i}, {j})"
            )
    else:
        return None


def get_closed_loop_system(pdata: PData, i: int = -1, j: int = -1) -> Any:
    """
    Returns the closed loop system in SciPy discrete LTI system form.

    :arg pdata dict: Problem data structure.
    :arg i int: Controller output index for the MIMO controller.
    :arg j int: Controller input index for the MIMO controller.

    :return scipy.signal.lti: SciPy (discrete) LTI system representation.

    Returns an `m` by `r` Python array when `i` or `j` is not provided.
    The ith row and jth column of the return value gives the discrete LTI
    system from jth input to the ith output.
    """
    if i == -1 or j == -1:
        return pdata["closed_loop_lti"]
    elif isinstance(pdata["closed_loop_lti"], list):
        try:
            return pdata["closed_loop_lti"][i][j]
        except IndexError:
            m = len(pdata["closed_loop_lti"])
            if m > 0 and isinstance(pdata["closed_loop_lti"][0], list):
                n = len(pdata["closed_loop_lti"][0])
            else:
                return None
            raise FocontError(
                f"Closed loop system dimension is {m}x{n}." f"Can not access ({i}, {j})"
            )
    else:
        return None


def norm(pdata: PData, cl: bool = True) -> float:
    r"""
    Calculates $\mathcal{H}_2$ norm of the closed or open loop
    MIMO system.

    :arg pdata dict: Problem data structure.
    :arg cl object: Calculate closed loop norm if it is `True`.

    :return float: $\mathcal{H}_2$ norm.
    """
    result: float = np.inf

    if cl:
        result = h2_norm(pdata["closed_loop_lti"])
    else:
        result = h2_norm(pdata["open_loop_lti"])

    return result


def h2_improvement(pdata: PData) -> float:
    r"""
    Compares the $\mathcal{H}_2$ norms of the closed loop
    system obtained by the algortihm and the open loop system
    if the open loop system is also stable.

    :arg pdata dict: Problem data structure.

    :return float: Ratio of the closed and open loop $\mathcal{H}_2$ norms.
    """
    ol_stable: bool = pdata["open_loop_stable"]

    if not ol_stable:
        warning(
            "Open loop system is not stable. " "Can not calculate H2 norm improvement."
        )

        return 0.0

    ol_h2: float = norm(pdata, False)
    print("Open loop H2 norm: {}".format(ol_h2))

    cl_h2: float = norm(pdata, True)
    print("Closed loop H2 norm: {}".format(cl_h2))

    result: float = cl_h2 / ol_h2
    print("Improvement: {}".format(result))

    return result


def bode(
    pdata: PData, loop: str, N: int = 256, xscale: str = "log", i: int = -1, j: int = -1
) -> Tuple[FOCArray, FOCArray]:
    lti: Any
    if loop == "open":
        lti = pdata["open_loop_lti"]
    elif loop == "closed":
        lti = pdata["closed_loop_lti"]

    r: int = len(lti)
    m: int = len(lti[0])
    A: FOCArray = lti[0][0].A
    B: FOCArray = np.zeros((A.shape[0], m))
    C: FOCArray = np.zeros((r, A.shape[0]))
    D: FOCArray = np.zeros((r, m))

    for k in range(r):
        C[k : k + 1, :] = lti[k][0].C

        for l in range(m):
            D[k : k + 1, l : l + 1] = lti[k][l].D

    for l in range(m):
        B[:, l : l + 1] = lti[0][l].B

    if i >= 0:
        C = C[i : i + 1, :]
        D = D[i : i + 1, :]

    if j >= 0:
        B = B[:, j : j + 1]
        D = D[:, j : j + 1]

    return freq_response(A, B, C, D, N)



import os
import json

from focont import foc, system


TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def test_01_01():
    filepath = TEST_DIR + '/aircraft_model.json'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_02():
    filepath = TEST_DIR + '/aircraft_model.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_03():
    filepath = TEST_DIR + '/unstable_aircraft_model.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_04():
    filepath = TEST_DIR + '/unstable_dt_aircraft_model.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_05():
    filepath = TEST_DIR + '/test_model_01.json'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_05_01():
    filepath = TEST_DIR + '/test_model_01.json'
    with open(filepath, 'r') as fh:
        input_data = json.load(fh)

        pdata = system.load(input_data)
        foc.solve(pdata)
        foc.print_results(pdata)

        assert foc.h2_improvement(pdata) < 1.0


def test_01_06():
    filepath = TEST_DIR + '/test_model_01_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)


def test_01_06_01():
    filepath = TEST_DIR + '/test_model_01_01_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_07():
    filepath = TEST_DIR + '/test_model_02_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_08():
    filepath = TEST_DIR + '/test_model_03_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_08():
    filepath = TEST_DIR + '/test_model_04_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_09():
    filepath = TEST_DIR + '/test_model_05_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_10():
    filepath = TEST_DIR + '/test_model_06_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


def test_01_11():
    filepath = TEST_DIR + '/test_model_07_for_FO.mat'
    pdata = system.load(filepath)
    foc.solve(pdata)
    foc.print_results(pdata)

    assert foc.h2_improvement(pdata) < 1.0


if __name__ == "__main__":
        test_01_01()

